﻿
using System;
using System.IO;
using Org.BouncyCastle.Bcpg;
using Org.BouncyCastle.Bcpg.OpenPgp;
using Org.BouncyCastle.Crypto.Generators;
using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Security;
using SafeOrbit.Memory;
using Org.BouncyCastle.Bcpg.Sig;
using Org.BouncyCastle.Asn1.Ocsp;
using System.Security.Cryptography;
using ProtectedData = SecureData.Data.ProtectedData;
using System.Runtime.Serialization;
using System.Text;
using static System.Runtime.InteropServices.JavaScript.JSType;
using System.Security.Cryptography.X509Certificates;


namespace SecureData.Managers
{
    public class SafeOrbitPGPCryptoManager
    {
        public string PublicKeyFile { get; private set; }
        public string PrivateKeyFile { get; private set; }

        public SafeOrbitPGPCryptoManager(string publicKeyPath, string privateKeyPath)
        {
            PublicKeyFile = publicKeyPath;
            PrivateKeyFile = privateKeyPath;
        }

        public void GenerateKeys(string identity, string passphrase)
        {
            PgpKeyRingGenerator keyRingGenerator = CreatePGPKeyRingGenerator(identity, passphrase);

            using (Stream pubOut = File.Create(PublicKeyFile))
            using (Stream privOut = File.Create(PrivateKeyFile))
            {
                keyRingGenerator.GeneratePublicKeyRing().Encode(pubOut);
                keyRingGenerator.GenerateSecretKeyRing().Encode(privOut);
            }

            Console.WriteLine("✅ Clés PGP générées avec succès !");
        }

        public void EncryptPassword(SafeObject<ProtectedData> safeData, string encryptedFile)
        {
            using (Stream keyIn = File.OpenRead(PublicKeyFile))
            using (Stream outputStream = File.Create(encryptedFile))
            using (Stream inputStream = new MemoryStream(System.Text.Encoding.UTF8.GetBytes(safeData.Object.Password)))
            {
                PgpPublicKey pubKey = ReadPublicKey(keyIn);
                PgpEncryptedDataGenerator encryptor = new PgpEncryptedDataGenerator(SymmetricKeyAlgorithmTag.Aes256);
                encryptor.AddMethod(pubKey);

                using (Stream encryptedOut = encryptor.Open(outputStream, new byte[1 << 16]))
                {
                    inputStream.CopyTo(encryptedOut);
                }
            }

            Console.WriteLine("🔐 Mot de passe chiffré avec succès !");
        }

        public string? DecryptPassword(string encryptedFile, string passphrase)
        {
            using (Stream keyIn = File.OpenRead(PrivateKeyFile))
            using (Stream encryptedStream = File.OpenRead(encryptedFile))
            {
                PgpSecretKeyRingBundle secretKeyRingBundle = new PgpSecretKeyRingBundle(PgpUtilities.GetDecoderStream(keyIn));
                PgpObjectFactory pgpF = new PgpObjectFactory(PgpUtilities.GetDecoderStream(encryptedStream));

                PgpEncryptedDataList? encDataList = pgpF.NextPgpObject() as PgpEncryptedDataList;
                PgpPrivateKey privateKey = FindPrivateKey(secretKeyRingBundle, passphrase);

                if (encDataList?.Count == 1)
                {
                    using (Stream decryptedStream = ((PgpPublicKeyEncryptedData)encDataList[0]).GetDataStream(privateKey))
                    using (StreamReader reader = new StreamReader(decryptedStream))
                    {
                        return reader.ReadToEnd();
                    }
                }
                return null;
            }
        }


        public void SignData(SafeObject<ProtectedData> safeData, string signedFile, string password)
        {
            using (Stream keyIn = File.OpenRead(PrivateKeyFile))
            using (Stream inputStream = new MemoryStream(System.Text.Encoding.UTF8.GetBytes(safeData.Object.Passphrase)))
            using (Stream outputStream = File.Create(signedFile))
            using (MemoryStream signedOut = new MemoryStream())
            {
                PgpSecretKeyRingBundle secretKeyRingBundle = new PgpSecretKeyRingBundle(PgpUtilities.GetDecoderStream(keyIn));
                PgpPrivateKey privateKey = FindPrivateKey(secretKeyRingBundle, password);

                PgpSignatureGenerator sGen = new PgpSignatureGenerator(privateKey.PublicKeyPacket.Algorithm, HashAlgorithmTag.Sha256);
                sGen.InitSign(PgpSignature.BinaryDocument, privateKey);

                Console.WriteLine("🔍 Initializing signature...");
                sGen.GenerateOnePassVersion(false).Encode(signedOut);
                inputStream.CopyTo(signedOut);
                sGen.Generate().Encode(signedOut);

                Console.WriteLine("🔍 Copying to output file...");
                signedOut.Flush();
                signedOut.Position = 0;
                signedOut.CopyTo(outputStream);
                outputStream.Flush();
            }

            Console.WriteLine("✅ Données signées avec succès !");
        }


        private PgpKeyRingGenerator CreatePGPKeyRingGenerator(string identity, string passphrase)
        {
            int certificationLevel = PgpSignature.PositiveCertification;

            PgpKeyPair masterKeyPair = GeneratePGPKeyPair(PublicKeyAlgorithmTag.RsaGeneral, 2048);
            PgpKeyPair subKeyPair = GeneratePGPKeyPair(PublicKeyAlgorithmTag.RsaGeneral, 2048);

            // ✅ Creating the hashed signature packets
            PgpSignatureSubpacketGenerator hashedSubpacketGenerator = new PgpSignatureSubpacketGenerator();
            hashedSubpacketGenerator.SetKeyFlags(true, KeyFlags.SignData | KeyFlags.EncryptStorage);
            hashedSubpacketGenerator.SetPreferredHashAlgorithms(true, new int[] { (int)HashAlgorithmTag.Sha256, (int)HashAlgorithmTag.Sha512 });

            PgpSignatureSubpacketVector hashedPackets = hashedSubpacketGenerator.Generate();

            // ✅ Creating the unhashed signature packets (optional metadata)
            PgpSignatureSubpacketGenerator unhashedSubpacketGenerator = new PgpSignatureSubpacketGenerator();
            PgpSignatureSubpacketVector unhashedPackets = unhashedSubpacketGenerator.Generate();

            SecureRandom rand = new SecureRandom();

            return new PgpKeyRingGenerator(
                certificationLevel,
                masterKeyPair,
                identity,
                SymmetricKeyAlgorithmTag.Aes256,
                passphrase.ToCharArray(),
                true, // useSha1
                hashedPackets,
                unhashedPackets,
                rand
            );
        }

        private PgpKeyPair GeneratePGPKeyPair(PublicKeyAlgorithmTag algorithm, int keySize)
        {
            RsaKeyPairGenerator keyGen = new RsaKeyPairGenerator();
            keyGen.Init(new KeyGenerationParameters(new SecureRandom(), keySize));
            AsymmetricCipherKeyPair keyPair = keyGen.GenerateKeyPair();

            return new PgpKeyPair(algorithm, keyPair.Public, keyPair.Private, DateTime.UtcNow);
        }

        private PgpPublicKey ReadPublicKey(Stream inputStream)
        {
            PgpPublicKeyRingBundle keyRingBundle = new PgpPublicKeyRingBundle(PgpUtilities.GetDecoderStream(inputStream));
            foreach (PgpPublicKeyRing keyRing in keyRingBundle.GetKeyRings())
            {
                foreach (PgpPublicKey key in keyRing.GetPublicKeys())
                {
                    if (key.IsEncryptionKey)
                        return key;
                }
            }
            throw new ArgumentException("Clé publique non trouvée.");
        }
        private PgpSignature LoadPGPSignature(PgpPublicKey publicKey,string signatureFilePath)
        {
            using (Stream keyIn = File.OpenRead(signatureFilePath))
            {
                PgpObjectFactory factory = new PgpObjectFactory(PgpUtilities.GetDecoderStream(keyIn));
                object obj =factory.NextPgpObject();
              
              
                return null;
           
            }
        }
        public  bool VerifySignature(string signatureFilePath, SafeObject<ProtectedData> safeData)
        {
            byte[] data = System.Text.Encoding.UTF8.GetBytes(safeData.Object.Passphrase);
            using (Stream keyIn = File.OpenRead(PublicKeyFile))
            {
                PgpPublicKey publicKey= ReadPublicKey(keyIn);
                PgpSignature signature = LoadPGPSignature(publicKey,signatureFilePath);
                signature.InitVerify(publicKey);

                using (Stream dataStream = new MemoryStream(data))
                {
                    int ch;
                    while ((ch = dataStream.ReadByte()) >= 0)
                    {
                        signature.Update((byte)ch);
                    }
                }
                return signature.Verify();

            }

            return false;
        }

        private PgpPrivateKey FindPrivateKey(PgpSecretKeyRingBundle secretKeyRingBundle, string passphrase)
        {
            foreach (PgpSecretKeyRing keyRing in secretKeyRingBundle.GetKeyRings())
            {
                foreach (PgpSecretKey key in keyRing.GetSecretKeys())
                {
                    PgpPrivateKey privateKey = key.ExtractPrivateKey(passphrase.ToCharArray());
                    if (privateKey != null)
                        return privateKey;
                }
            }
            throw new ArgumentException("Clé privée non trouvée.");
        }
    }
}